/*
TV-B-Gone
Firmware
for use with ATtiny85
Mitch Altman + Limor Fried
7-Oct-07

Merged V1.0 board code in with V1.1 board code
using #defines. The operation of stealth mode is
consistent between both board versions, the opertion
of persistent stealth mode however, is different.

To properly operate persistent stealth on V1.0, you have
to push, release, then push and hold the button for
5 seconds.

On Version 1.1, you push the button 11 times, with less
than 2 seconds delay between each press.

The TV code database compression code was written by
Kevin Zimmerman.

Damien Good
7-Dec-07


Rewrote the Button multiclick handler.  Now the operation is consistent
on both Version 1.0 and Version 1.1 hardware.
The operation is as follows. (assuming persistent stealth is off.)

1 Click - Blink TV power codes, and blink LED
2 Clicks - Blink TV power codes, no LED blink.
3 Clicks - Blink the TV-B-Gone region programmed. (3 for NA/Asia, 6 for EU)
4 Clicks - Same as 1 Click.
5 Clicks - Turn Persistent Stealth mode on/off.

Also Rewrote to allow for pulse codes, as well as modulated codes.

Damien Good
8-Dec-07

Distributed under Creative Commons 2.5 -- Attib & Share Alike
*/

/*
Uncomment the following line to build the Version 1.0 board firmware, which
includes stealth mode and persistent stealth mode.
*/
//#define VERSION_1_0_BOARD
/*
Uncomment the following line to Test Version 1.1 board firmware on a Version
1.0 board.
*/
//#define VERSION_1_1_TEST_ON_1_0

/*
You can also define the above when doing make program.

The syntax for that, is
make program VER_10=1 (VERSION_1_0_BOARD defined)
or
make program VER_10_TEST=1 (VERSION_1_1_TEST_ON_1_0 defined)

If you wish to compile and program in European codes instead, then
it is:
make program EU_CODE=1

Otherwise, THe North America/Asia code database will be programmed
in by default.
*/

#include <avr/io.h>             // this contains all the IO port definitions
#ifdef VERSION_1_0_BOARD
	#include <avr/interrupt.h>      // definitions for interrupts
#endif
#include <avr/eeprom.h>
#include <avr/sleep.h>          // definitions for power-down modes
#include <avr/pgmspace.h>       // definitions or keeping constants in program memory
#include "main.h"



						
#ifdef VERSION_1_0_BOARD
  #define LED PB0         //  visible LED
  #define BUTTON PB2      //  IR LED
  #define IRLED PB1      //  IR LED
#else
  #ifdef VERSION_1_1_TEST_ON_1_0
    #define LED PB0
    #define IRLED1 PB2
  #else
    #define LED PB2
    #define IRLED1 PB0
  #endif
    #define IRLED2 PB1
#endif
	
#ifdef NA_CODES
	#define NUMBER_BLINKS 3
#else
	#define NUMBER_BLINKS 6
#endif

/*
This project transmits a bunch of TV POWER codes, one right after the other, 
with a pause in between each.  (To have a visible indication that it is 
transmitting, it also pulses a visible LED once each time a POWER code is 
transmitted.)  That is all TV-B-Gone does.  The tricky part of TV-B-Gone 
was collecting all of the POWER codes, and getting rid of the duplicates and 
near-duplicates (because if there is a duplicate, then one POWER code will 
turn a TV off, and the duplicate will turn it on again (which we certainly 
do not want).  I have compiled the top-40 most popular codes with the 
duplicates eliminated, both for North America (which is the same as Asia, as 
far as POWER codes are concerned -- even though much of Asia USES PAL video) 
and for Europe (which works for Australia, New Zealand, the Middle East, and 
other parts of the world that use PAL video).

Before creating a TV-B-Gone Kit, I originally started this project by hacking 
the MiniPOV kit.  This presents a limitation, based on the size of
the Atmel ATtiny2313 internal flash memory, which is 2KB.  With 2KB we can only 
fit about 7 POWER codes into the firmware's database of POWER codes.  40 codes
requires 8KB of flash memory, which is why we chose the ATtiny85 for the 
TV-B-Gone Kit.

This version of the firmware has the most popular 40 POWER codes for North America.
*/


/*
This project is a good example of how to use the AVR chip timers.
*/


/*
The hardware for this project is very simple:
     ATtiny85 has 8 pins:
       pin 1   connects to programming circuitry
       pin 2   one pin of ceramic resonator
       pin 3   one pin of ceramic resonator
       pin 4   ground
       pin 5   PB0 - visible LED, and also connects to programming circuitry
       pin 6   OC1A - IR emitter, through a PN2222A driver (with 47 ohm base resistor), and also connects to programming circuitry
       pin 7   push-button switch, and also connects to serial port programming circuitry
       pin 8   +3v
    See the schematic for more details.

    This firmware requires using an 8.0MHz ceramic resonator 
       (since the internal oscillator may not be accurate enough).

    IMPORTANT:  to use the ceramic resonator, you must perform the following:
                    make burn-fuse_cr
*/


/*
The C compiler creates code that will transfer all constants into RAM when the microcontroller
resets.  Since this firmware has a table (powerCodes) that is too large to transfer into RAM,
the C compiler needs to be told to keep it in program memory space.  This is accomplished by
the macro PROGMEM (this is used in the definition for powerCodes).  Since the
C compiler assumes that constants are in RAM, rather than in program memory, when accessing
powerCodes, we need to use the pgm_read_word() and pgm_read_byte macros, and we need
to use powerCodes as an address.  This is done with PGM_P, defined below.  
For example, when we start a new powerCode, we first point to it with the following statement:
    PGM_P thecode_p = pgm_read_word(powerCodes+i);
The next read from the powerCode is a byte that indicates the carrier frequency, read as follows:
    uint8_t freq = pgm_read_byte(thecode_p);
Subsequent reads from the powerCode are onTime/offTime pairs, which are words, read as follows:
    ontime = pgm_read_word(thecode_p+(offset_into_table);
    offtime = pgm_read_word(thecode_p+(offset_into_table);
*/

#ifdef VERSION_1_0_BOARD
SIGNAL(SIG_INTERRUPT0) {
  PORTB &= ~_BV(LED); // turn on visible LED at PB0 by pulling pin to ground
  TCCR1 = 0;          // turn off Timer1 frequency generator (should be off already)
  PORTB &= ~_BV(IRLED);            // IR LED is off when pin is low
  WDTCR = _BV(WDE);   // turn on the watchdog timer to reset the chip
}
#endif

#define NOP __asm__ __volatile__ ("nop")
// This function delays the specified number of 10 microseconds
#define DELAY_CNT 11
void delay_ten_us(uint16_t us) {
  uint8_t timer;
  while (us != 0) {
    for (timer=0; timer <= DELAY_CNT; timer++) {
      NOP;
      NOP;
    }
    NOP;
    us--;
  }
}


// This function quickly pulses the visible LED (connected to PB0, pin 5)
void quickflashLED( void ) {
  // pulse LED on for 30ms

  PORTB &= ~_BV(LED);         // turn on visible LED at PB0 by pulling pin to ground
  delay_ten_us(3000);         // 30 millisec delay
  PORTB |= _BV(LED);          // turn off visible LED at PB0 by pulling pin to +3V
}

void quickflashLEDx( uint8_t x ) {
  quickflashLED();
  while(--x) {
	delay_ten_us(15000);        // 150 millisec delay
	quickflashLED();
  }
}

// This function transmits one Code Element of a POWER code to the IR emitter, 
//   given offTime and onTime for the codeElement
//     If offTime = 0 that signifies the last Code Element of the POWER code
//     and the delay_ten_us function will have no delay for offTime 
//     (but we'll delay for 250 milliseconds in the main function)
#ifndef VERSION_1_0_BOARD
void xmitCodeElement(uint16_t ontime, uint16_t offtime, uint8_t pulse_code ) {
  // start Timer1 outputting the carrier frequency to IR emitters on OC1A (PB1, pin 6) and OC0A (PB0, pin 5)
  TCNT0 = 0; // reset the timers so they are aligned
  TCNT1 = 0;
  TIFR = 0;  // clean out the timer flags
  if(!pulse_code)
  {
	#ifndef VERSION_1_1_TEST_ON_1_0
	  TCCR0A =_BV(COM0A0) | _BV(WGM01);          // set up timer 0
	#endif

	  TCCR1 =_BV(COM1A0) | _BV(CS10) | _BV(CTC1);      // set up and turn on timer 1
	#ifndef VERSION_1_1_TEST_ON_1_0
	  TCCR1 = 0b10010001;  // CTC1 = 1 to reset Timer1 to 0 when it reaches the value in OCR1C (i.e., on Compare Match)
	                        // PWM1A = 0 to disable PWM mode
	                        // COM1A1:0 = 01 to toggle OC1A on Compare Match
	                        // CS13:10 = 0001 to start Timer1 with prescaler set to divide by 1 (i.e., no prescaler divide)
	  TCCR0B = _BV(CS00);                              // turn on timer 0 exactly 1 instruction later
	#endif
  }
  else
  {
  	  #ifndef VERSION_1_1_TEST_ON_1_0
  	  PORTB |= _BV(IRLED1) | _BV(IRLED2);
  	  #else
  	  PORTB |= _BV(IRLED2);
  	  #endif
  }

  // keep transmitting carrier for onTime
  delay_ten_us(ontime);

  
  // turn off output to IR emitters on 0C1A (PB1, pin 6) for offTime
  TCCR1 = 0;      // stop Timer 1
  TCCR0B = 0;     // stop Timer 0, exactly one instruction later
  TCCR0A = 0;

  PORTB &= ~_BV(IRLED1) & ~_BV(IRLED2);           // turn off IR LED

  delay_ten_us(offtime);
}
#else
void xmitCodeElement(uint16_t ontime, uint16_t offtime, uint8_t pulse_code ) {
  // start Timer1 outputting the carrier frequency to IR emitters on OC1A (PB1, pin 6)
  if(!pulse_code)
    TCCR1 =_BV(COM1A0) | _BV(CS10) | _BV(CTC1);
  else
  	PORTB |= _BV(IRLED);
  //TCCR1 = 0b10010001  // CTC1 = 1 to reset Timer1 to 0 when it reaches the value in OCR1C (i.e., on Compare Match)
                        // PWM1A = 0 to disable PWM mode
                        // COM1A1:0 = 01 to toggle OC1A on Compare Match
                        // CS13:10 = 0001 to start Timer1 with prescaler set to divide by 1 (i.e., no prescaler divide)
 
  //PORTB &= ~_BV(LED);
  // keep transmitting carrier for onTime
  delay_ten_us(ontime);

  // turn off output to IR emitters on 0C1A (PB1, pin 6) for offTime
  TCCR1 = 0;      // stop Timer1
  PORTB &= ~_BV(IRLED);  // turn off IR LED
  //PORTB |= _BV(LED);
  delay_ten_us(offtime);
}
#endif

//extern const struct powercode powerCodes[] PROGMEM;
extern const PGM_P *powerCodes[] PROGMEM;

extern uint8_t num_codes;

//#ifdef VERSION_1_0_BOARD
uint16_t EEMEM eeprom_initialized = 0x6502;
uint8_t EEMEM eeprom_disableblink;
//#endif
static volatile int KeepTrack __attribute__ ((section (".noinit")));

uint8_t bitsleft_r = 0;
uint8_t bits_r=0;
PGM_P code_ptr;

uint8_t read_bits(uint8_t count)
{
	uint8_t i;
	uint8_t tmp=0;

	for(i=0;i<count;i++)
	{
		if(bitsleft_r==0)
		{
			bits_r = pgm_read_byte(code_ptr++);
			bitsleft_r = 8;
		}
		bitsleft_r--;
		tmp |= (((bits_r >> (bitsleft_r)) & 1) << (count-1-i));
	}
	return tmp;

}

void tvbgone_sleep( void )
{
#ifndef VERSION_1_0_BOARD
  // Shut down everything and put the CPU to sleep
  TCCR1 = 0;                      // turn off frequency generator (should be off already)
  TCCR0B = 0;
  PORTB |= _BV(LED); // turn on the button pullup, turn off visible LED
  PORTB &= ~_BV(IRLED1) & ~_BV(IRLED2);           // turn off IR LED
#else
  TCCR1 = 0;                      // turn off frequency generator (should be off already)
  PORTB = _BV(BUTTON) | _BV(LED); // turn on the button pullup, turn off visible LED
  PORTB &= ~_BV(IRLED);           // turn off IR LED
#endif
  delay_ten_us(1000);             // wait 10 millisec second

  MCUCR = _BV(SM1) |  _BV(SE);    // power down mode,  SE=1 (bit 5) -- enables Sleep Modes
  sleep_cpu();                    // put CPU inteo Power Down Sleep Mode
}

void flashslowLEDx ( uint8_t enable_flash, uint8_t num_blinks )
{
	uint8_t i;
	for(i=0;i<num_blinks;i++)
	{
		if(enable_flash)
			PORTB &= ~_BV(LED);         // turn on visible LED at PB0 by pulling pin to ground
	  	delay_ten_us(50000);         // 500 millisec delay
	  	PORTB |= _BV(LED);          // turn off visible LED at PB0 by pulling pin to +3V
	  	delay_ten_us(50000);		// 500 millisec delay
  	}
}

int main(void) {
  uint16_t ontime, offtime;
  uint8_t i, Loop;
#ifndef VERSION_1_0_BOARD
  DDRB = _BV(LED) | _BV(IRLED1) | _BV(IRLED2);    // set the visible and IR LED pins to outputs
  PORTB = _BV(LED);  //  visible LED is off when pin is high
  PORTB &= ~_BV(IRLED1) & ~_BV(IRLED2);        // IR LED is off when pin is low
  
  
#else
	
  MCUSR = 0;                     // clear watchdog flag
  WDTCR = _BV(WDCE) | _BV(WDE);  // enable WDT disable
  WDTCR = 0;                     // disable WDT

  GIMSK = _BV(INT0);     // generate an interrupt on low level of INT0
  TIMSK = 0x00;          // no Timer interrupts enabled
  sei();
  
  DDRB = _BV(LED) | _BV(IRLED);    // set the visible and IR LED pins to outputs
  PORTB = _BV(BUTTON) | _BV(LED);  // button has a pullup, visible LED is off when pin is high
  PORTB &= ~_BV(IRLED);            // IR LED is off when pin is low
#endif
  
  if(eeprom_read_word(&eeprom_initialized)!=0x6502)	//Make sure eeprom is initialized to default values.
  {
  	  eeprom_write_word(&eeprom_initialized,0x6502);
  	  eeprom_write_byte(&eeprom_disableblink,0);
  }
  
  
  uint8_t disable_blink = eeprom_read_byte(&eeprom_disableblink);  //Flag to disable blink.
  delay_ten_us(5000);   //Prevent switch bounce being detected.
  
  KeepTrack++;	//KeepTrack is Declared as a static volatile integer, that is NOT initialized on reset.
  				//As such, can be used to track the number of times the button has been pressed.
  if((KeepTrack-1)>5)
  	  KeepTrack=1;
  flashslowLEDx(1,KeepTrack); //To indicate which user mode.
  flashslowLEDx(0,1);	//To give the user time to make next selection
  switch(KeepTrack-1)
  {
  	default:
  		Loop = 0;
  		break;
   	case 1:
  		disable_blink = !disable_blink;
  		Loop = 0;
   		break;
   	case 2:
   		Loop = 1;	//Enable persistent stealth, if you want a stealth loop.
   		break;		//Single click to restart loop, double_click to exit loop.
   	case 3:
   		flashslowLEDx(1,NUMBER_BLINKS);
   		KeepTrack=0;
   		tvbgone_sleep();
   		break;
   	case 4:
   		flashslowLEDx(disable_blink,NUMBER_BLINKS);
   		KeepTrack=0;
   		tvbgone_sleep();
   		break;
   	case 5:
   		disable_blink = !disable_blink;
		eeprom_write_byte(&eeprom_disableblink,disable_blink);
  		quickflashLEDx(8);
  		KeepTrack=0;
  		tvbgone_sleep();
  		break;
  }
	
  if(Loop==0)
  	KeepTrack=0;
  else
  	KeepTrack--;	//To allow restart loop at beginning.  
  
do {	//Execute the code at least once.  If Loop is on, execute forever.
  for(i=0; i<num_codes; i++) {   // for every POWER code in our collection

    

    code_ptr = (PGM_P)pgm_read_word(powerCodes+i);     // point to next POWER code
    //DEBUG(putstring("\n\rAddr: "); putnum_uh(thecode_p));

	// get carrier period
    const uint8_t freq = pgm_read_byte(code_ptr++);
    // set OCR for Timer1 to output this POWER code's carrier frequency
    OCR1C = freq; 

    //uint16_t x = (freq+1) * 2;
    //DEBUG(putstring("\n\rOCR1: "); putnum_ud(freq););
    //DEBUG(putstring("\n\rFreq: "); putnum_ud(F_CPU/x););

	// get pointer to pulse times array
	const PGM_P time_ptr = (PGM_P)pgm_read_word(code_ptr);
	code_ptr+=2;
	//const com_type = pgm_read_byte(code_ptr);
	//code_ptr++;
	//uint8_t com_type=read_bits(2);
	//if(com_type==0)
	//	read_bits(2);
	uint8_t com_type=0;

    // transmit all codeElements for this POWER code (a codeElement is an onTime and an offTime)
    // transmitting onTime means pulsing the IR emitters at the carrier frequency for the length of time specified in onTime
    // transmitting offTime means no output from the IR emitters for the length of time specified in offTime
    do {
	  uint8_t ti;
	  ti = ( read_bits(4-com_type) << 2 );
	 

      // read the onTime and offTime from the program memory
	  ontime = pgm_read_word(time_ptr+ti);
	  offtime = pgm_read_word(time_ptr+ti+2);
      //DEBUG(putstring("\n\r\tPair = "); putnum_ud(ontime));
      //DEBUG(putstring("\t"); putnum_ud(offtime));

      xmitCodeElement(ontime, offtime, (freq==0));  // transmit this codeElement (ontime and offtime)
    } while ( offtime != 0 );  // offTime = 0 signifies last codeElement for a POWER code

    // delay 250 milliseconds before transmitting next POWER code
    bitsleft_r=0;	//Flush remaining bits, so that next code starts
    				//with a fresh set of 8 bits.
    delay_ten_us(25000);
//#ifdef VERSION_1_0_BOARD
	if(!disable_blink)
//#endif
    quickflashLED(); // visible indication that a code has been output.
    	
  }
} while (Loop == 1);
  

  // flash the visible LED on PB0  4 times to indicate that we're done
  delay_ten_us(65500); // wait maxtime 
  quickflashLEDx(4);

  tvbgone_sleep();
}

